/******************************************************************************
  Copyright 2013 Tom Cornall

  This program is free software: you can redistribute it and/or modify
  it under the terms of the GNU General Public License as published by
  the Free Software Foundation, either version 3 of the License, or
  (at your option) any later version.

  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.

  You should have received a copy of the GNU General Public License
  along with this program.  If not, see <http://www.gnu.org/licenses/>.  
 ******************************************************************************/
#include <LUFA/Drivers/USB/USB.h>
#include <avr/interrupt.h>
#include <avr/io.h>
#include <avr/power.h>
#include <avr/wdt.h>
#include <string.h>
#include "descriptors.h"
#include "diag.h"
#include "kbd.h"
#include "sr.h"
#include "vref.h"

static uint8_t prevKeyboardHIDReportBuf[sizeof(USB_KeyboardReport_Data_t)];
static uint8_t prevGenericHIDReportBuf[GENERIC_REPORT_SIZE];

USB_ClassInfo_HID_Device_t keyboardHIDIface = {
    .Config = {
        .InterfaceNumber = 0,
        .ReportINEndpoint = {
            .Address = KEYBOARD_IN_EPADDR,
            .Size = HID_EPSIZE,
            .Banks = 1,
        },
        .PrevReportINBuffer = prevKeyboardHIDReportBuf,
        .PrevReportINBufferSize = sizeof(prevKeyboardHIDReportBuf),
    }
};

USB_ClassInfo_HID_Device_t genericHIDIface = {
    .Config = {
        .InterfaceNumber = 1,
        .ReportINEndpoint = {
            .Address = GENERIC_IN_EPADDR,
            .Size = HID_EPSIZE,
            .Banks = 1,
        },
        .PrevReportINBuffer = prevGenericHIDReportBuf,
        .PrevReportINBufferSize = sizeof(prevGenericHIDReportBuf),
    }
};

/*
 * Event handlers
 */
void EVENT_USB_Device_ConfigurationChanged(void)
{
    HID_Device_ConfigureEndpoints(&keyboardHIDIface);
    HID_Device_ConfigureEndpoints(&genericHIDIface);
    USB_Device_EnableSOFEvents();
}
void EVENT_USB_Device_ControlRequest(void)
{
    HID_Device_ProcessControlRequest(&keyboardHIDIface);
    HID_Device_ProcessControlRequest(&genericHIDIface);
}
void EVENT_USB_Device_StartOfFrame(void)
{
    HID_Device_MillisecondElapsed(&keyboardHIDIface);
    HID_Device_MillisecondElapsed(&genericHIDIface);
}

/*
 * Send keyboard state to host
 */
bool CALLBACK_HID_Device_CreateHIDReport(
        USB_ClassInfo_HID_Device_t *const hidIfaceInfo, uint8_t *const reportID,
        const uint8_t reportType, void *reportData, uint16_t *const reportSize)
{
    if (hidIfaceInfo == &keyboardHIDIface)
    {
        USB_KeyboardReport_Data_t *kbdReport =
            (USB_KeyboardReport_Data_t *)reportData;

        kbdFillReport(kbdReport);

        *reportSize = sizeof(USB_KeyboardReport_Data_t);
        return false;
    }
    else
    {
        diagFillReport(reportData);

        *reportSize = GENERIC_REPORT_SIZE;
        return false;
    }
}

/*
 *
 */
void CALLBACK_HID_Device_ProcessHIDReport(
        USB_ClassInfo_HID_Device_t *const hidIfaceInfo, const uint8_t reportID,
        const uint8_t reportType, const void *reportData,
        const uint16_t reportSize)
{
    if (hidIfaceInfo == &genericHIDIface)
        diagReceiveReport(reportData);
}

/*
 *
 */
void initHardware(void)
{
    /* disable watchdog */
    MCUSR &= ~(1 << WDRF);
    wdt_disable();

    /* clock full speed */
    clock_prescale_set(clock_div_1);

    /* Hardware Initialization */
    vrefInit();
    srInit();
    srClear();
    srEnable();
    USB_Init();
}

/*
 *
 */
int main(void)
{
    initHardware();
    sei();

    layersLoad();

    _delay_ms(5);

    /* try loading Vref from eeprom; if set to 0xffff, run calibration */
    vrefLoad();
    if (vrefGet() == 0xffff)
        kbdCalibrate();

    for (;;)
    {
        kbdScan();
        HID_Device_USBTask(&keyboardHIDIface);
        HID_Device_USBTask(&genericHIDIface);
        USB_USBTask();
    }
}
