/******************************************************************************
  Copyright 2013 Tom Cornall

  This program is free software: you can redistribute it and/or modify
  it under the terms of the GNU General Public License as published by
  the Free Software Foundation, either version 3 of the License, or
  (at your option) any later version.

  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.

  You should have received a copy of the GNU General Public License
  along with this program.  If not, see <http://www.gnu.org/licenses/>.  
 ******************************************************************************/
#include "diag.h"

#if defined (__AVR_ATmega8U2__)
void (*bootloader)(void) = (void *) 0x0800;
#elif defined (__AVR_ATmega32U2__)
void (*bootloader)(void) = (void *) 0x1800;
#endif

static DiagReportState diagReportState = dr1stCols;

static void diagReportKbd(uint8_t *reportData, uint8_t startCol,
        uint8_t endCol);
static void diagReportVref(uint8_t *reportData);
static void diagReportScanCodes(uint8_t *reportData);
static void diagReportLayerConditions(uint8_t *reportData);
static void diagSetScanCode(uint8_t layer, uint8_t col, uint8_t row,
        uint8_t sc);
static void diagSetLayerCondition(uint8_t conditionIdx, uint8_t val);

/*
 *
 */
void diagReceiveReport(const uint8_t *reportData)
{
    DiagReportCommand cmd = reportData[0];

    switch (cmd)
    {
        case dcSetState:
            diagReportState = reportData[1];
            break;
        case dcSetVref:
            vrefSet(reportData[1] + (reportData[2] << 8));
            break;
        case dcJumpBootloader:
            if (reportData[1])
                bootloader();
            break;
        case dcStoreVref:
            if (reportData[1])
                vrefStore();
            break;
        case dcAutoCalibrate:
            if (reportData[1])
                kbdCalibrate();
            break;
        case dcSetScanCode:
            diagSetScanCode(reportData[1] >> 6, reportData[1] & 0x3f,
                    reportData[2], reportData[3]);
            break;
        case dcSetLayerCondition:
            diagSetLayerCondition(reportData[1], reportData[2]);
            break;
        case dcLoadScanCodes:
            if (reportData[1])
                layersLoad();
            break;
        case dcStoreScanCodes:
            if (reportData[1])
                layersStore();
            break;
    }
}

/*
 *
 */
void diagFillReport(uint8_t *reportData)
{
    /* zero report */
    memset(reportData, 0x00, GENERIC_REPORT_SIZE);

    /* first byte of report is reporting state */
    reportData[0] = diagReportState;

    switch (diagReportState)
    {
        case dr1stCols:
            diagReportKbd(reportData, 0, 11);
            diagReportState = dr2ndCols;
            break;
        case dr2ndCols:
            diagReportKbd(reportData, 12, 22);
            diagReportState = dr1stCols;
            break;
        case drVref:
            diagReportVref(reportData);
            break;
        case drScanCodes:
            diagReportScanCodes(reportData);
            break;
        case drLayerConditions:
            diagReportLayerConditions(reportData);
            break;
        default:
            break;
    }
}

/*
 *
 */
void diagReportKbd(uint8_t *reportData, uint8_t startCol, uint8_t endCol)
{
    /* data is sent column-major, so the first data byte looks like:
     *
     *     b7     b6     b5    b4     b3     b2     b1     b0
     *   r3(c1).r2(c1).r1(c1).r0(c0).r3(c0).r2(c0).r1(c0).r0(c0)
     */
    uint8_t currByte = 1;
    uint8_t col = startCol;
    while (col <= endCol)
    {
        if (kbd[col][0] > KBD_DEBOUNCE_THRESH)
            reportData[currByte] |= (1 << 0);
        if (kbd[col][1] > KBD_DEBOUNCE_THRESH)
            reportData[currByte] |= (1 << 1);
        if (kbd[col][2] > KBD_DEBOUNCE_THRESH)
            reportData[currByte] |= (1 << 2);
        if (kbd[col][3] > KBD_DEBOUNCE_THRESH)
            reportData[currByte] |= (1 << 3);

        col++;
        if (col > endCol)
            break;

        if (kbd[col][0] > KBD_DEBOUNCE_THRESH)
            reportData[currByte] |= (1 << 4);
        if (kbd[col][1] > KBD_DEBOUNCE_THRESH)
            reportData[currByte] |= (1 << 5);
        if (kbd[col][2] > KBD_DEBOUNCE_THRESH)
            reportData[currByte] |= (1 << 6);
        if (kbd[col][3] > KBD_DEBOUNCE_THRESH)
            reportData[currByte] |= (1 << 7);

        col++;
        currByte++;
    }
}

/*
 *
 */
void diagReportVref(uint8_t *reportData)
{
    uint16_t vref = vrefGet();
    reportData[1] = vref & 0xff;
    reportData[2] = vref >> 8;
}

/*
 *
 */
void diagReportScanCodes(uint8_t *reportData)
{
    static uint8_t currLayer = 0;
    static uint8_t (*currMtx)[KBD_COLS][KBD_ROWS] = &layerMatrix0;
    static uint8_t currCol = 0;

    /* first byte of report has the layer in the top two bits, with the column
     * in the bottom 6.
     */
    reportData[1] = (currLayer << 6) | currCol;

    for (uint8_t i = 0; i < KBD_ROWS; i++)
        reportData[i + 2] = (*currMtx)[currCol][i];

    currCol++;
    if (currCol == 23)
    {
        currCol = 0;

        currLayer++;
        if (currLayer == 4)
            currLayer = 0;
        switch (currLayer)
        {
            case 0:
                currMtx = &layerMatrix0;
                break;
            case 1:
                currMtx = &layerMatrix1;
                break;
            case 2:
                currMtx = &layerMatrix2;
                break;
            case 3:
                currMtx = &layerMatrix3;
                break;
        }
    }
}

/*
 *
 */
void diagReportLayerConditions(uint8_t *reportData)
{
    static uint8_t currSet = 0; // sent in two halves

    reportData[1] = currSet;

    memcpy(&reportData[2],
            &layerConditions[currSet * (LAYERS_NUM_CONDITIONS / 2)],
            LAYERS_NUM_CONDITIONS / 2);

    currSet++;
    if (currSet == 2)
        currSet = 0;
}

/*
 *
 */
void diagSetScanCode(uint8_t layer, uint8_t col, uint8_t row, uint8_t sc)
{
    if (col >= 0 && col < KBD_COLS && row >= 0 && row < KBD_ROWS)
        (*layersMatrix(layer))[col][row] = sc;
}

/*
 *
 */
void diagSetLayerCondition(uint8_t conditionIdx, uint8_t val)
{
    if (conditionIdx >= 0 && conditionIdx < LAYERS_NUM_CONDITIONS)
        layerConditions[conditionIdx] = val;
}
