/******************************************************************************
  Copyright 2013 Tom Cornall

  This program is free software: you can redistribute it and/or modify
  it under the terms of the GNU General Public License as published by
  the Free Software Foundation, either version 3 of the License, or
  (at your option) any later version.

  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.

  You should have received a copy of the GNU General Public License
  along with this program.  If not, see <http://www.gnu.org/licenses/>.  
 ******************************************************************************/
#include "kbd.h"

int8_t kbd[KBD_COLS][KBD_ROWS];
uint8_t kbdBitmap[(KBD_ROWS * KBD_COLS + 8 - 1) / 8];

static void kbdScanColumn(uint8_t col);

/*
 *
 */
void kbdCalibrate(void)
{
    uint16_t rMin, rMax, val = 0;
    uint16_t loPoint, hiPoint;

    uint8_t numLO = 0;
    uint8_t numHI = 0;

    for (uint8_t col = 0; col < KBD_COLS; col++)
    {
        for (uint8_t row = 0; row < KBD_ROWS; row++)
        {
            switch (layerMatrix0[col][row])
            {
                case KBD_SC_CAL_HI:
                    numHI++;
                    break;
                case KBD_SC_CAL_LO:
                    numLO++;
                    break;
            }
        }
    }

    /* first look point where all LO keys first start staying LO */
    rMin = 0;
    rMax = VREF_MAX;
    while (rMax - rMin > 1)
    {
        val = ((rMax - rMin) / 2) + rMin;
        vrefSet(val);
        _delay_us(100); // wait for DAC to settle

        for (uint8_t i = 0; i < 32; i++)
            kbdScan();

        uint8_t pressedLoKeys = 0;
        for (uint8_t col = 0; col < KBD_COLS; col++)
            for (uint8_t row = 0; row < KBD_ROWS; row++)
                if (kbd[col][row] > KBD_DEBOUNCE_THRESH &&
                        layerMatrix0[col][row] >= HID_KEYBOARD_SC_A)
                    pressedLoKeys++;

        if (pressedLoKeys < numLO)
            rMin = val;
        else
            rMax = val;
    }
    loPoint = val;

    /* now look for point where all HI keys first start staying LO */
    rMin = 0;
    rMax = VREF_MAX;
    while (rMax - rMin > 1)
    {
        val = ((rMax - rMin) / 2) + rMin;
        vrefSet(val);
        _delay_us(100); // wait for DAC to settle

        for (uint8_t i = 0; i < 32; i++)
            kbdScan();

        uint8_t pressedHiKeys = 0;
        for (uint8_t col = 0; col < KBD_COLS; col++)
            for (uint8_t row = 0; row < KBD_ROWS; row++)
                if (kbd[col][row] > KBD_DEBOUNCE_THRESH &&
                        layerMatrix0[col][row] == KBD_SC_CAL_HI)
                    pressedHiKeys++;

        if (pressedHiKeys < numHI)
            rMin = val;
        else
            rMax = val;
    }
    hiPoint = val;

    /* set 3/4 of the way between two points. wild stab in the dark. */
    vrefSet(((loPoint - hiPoint) / 4) * 3 + hiPoint);
}

/*
 *
 */
void kbdFillReport(USB_KeyboardReport_Data_t *kbdReport)
{
    uint8_t usedKeyCodes = 0;

    uint8_t (*mtx)[KBD_COLS][KBD_ROWS] = layersMatrix(layersWhichLayer());

    /* debug output from PROG pushbutton */
    if (!(PIND &= (1 << PD7)))
        kbdReport->KeyCode[usedKeyCodes++] = HID_KEYBOARD_SC_P;

    for (uint8_t col = 0; col < KBD_COLS; col++)
    {
        for (uint8_t row = 0; row < KBD_ROWS && usedKeyCodes < 6; row++)
        {
            if (kbdBitmap[KBD_BMP_BYTE(col)] & KBD_BMP_MASK(col, row))
            {
                uint8_t sc = (*mtx)[col][row];

                if (sc < HID_KEYBOARD_SC_A ||
                        (sc >= KBD_SC_FN1 && sc <= KBD_SC_FN3))
                    continue;

                kbdReport->KeyCode[usedKeyCodes++] = sc;
                switch (sc)
                {
                    case HID_KEYBOARD_SC_LEFT_SHIFT:
                        kbdReport->Modifier +=
                            HID_KEYBOARD_MODIFIER_LEFTSHIFT;
                        break;
                    case HID_KEYBOARD_SC_RIGHT_SHIFT:
                        kbdReport->Modifier +=
                            HID_KEYBOARD_MODIFIER_RIGHTSHIFT;
                        break;
                    case HID_KEYBOARD_SC_LEFT_CONTROL:
                        kbdReport->Modifier +=
                            HID_KEYBOARD_MODIFIER_LEFTCTRL;
                        break;
                    case HID_KEYBOARD_SC_RIGHT_CONTROL:
                        kbdReport->Modifier +=
                            HID_KEYBOARD_MODIFIER_RIGHTCTRL;
                        break;
                    case HID_KEYBOARD_SC_LEFT_ALT:
                        kbdReport->Modifier +=
                            HID_KEYBOARD_MODIFIER_LEFTALT;
                        break;
                    case HID_KEYBOARD_SC_RIGHT_ALT:
                        kbdReport->Modifier +=
                            HID_KEYBOARD_MODIFIER_RIGHTALT;
                        break;
                    case HID_KEYBOARD_SC_LEFT_GUI:
                        kbdReport->Modifier +=
                            HID_KEYBOARD_MODIFIER_LEFTGUI;
                        break;
                    case HID_KEYBOARD_SC_RIGHT_GUI:
                        kbdReport->Modifier +=
                            HID_KEYBOARD_MODIFIER_RIGHTGUI;
                        break;
                }
            }
        }
    }
}

/*
 *
 */
void kbdScan(void)
{
    memset(kbdBitmap, 0x00, sizeof(kbdBitmap));

    for (uint8_t i = 0; i < KBD_COLS; i++)
    {
        /* select column */
        srDisable();
        srSet(i);
        srEnable();

        /* scan */
        kbdScanColumn(i);

        /* deselect column */
        srDisable();
        srClear();
        srEnable();

        /* wait for negative spike */
        _delay_loop_1(50);
    }
}

/*
 *
 */
void kbdScanColumn(uint8_t col)
{
    uint8_t scan[KBD_ROWS];
    memset(scan, 0x00, KBD_ROWS);

    _delay_loop_1(10); // LM339 propagation delay
    if (KBD_ROWS_PIN & (1 << KBD_ROW1_IP)) // 1st row
        scan[0] = 1;
    if (KBD_ROWS_PIN & (1 << KBD_ROW2_IP)) // 2nd row
        scan[1] = 1;
    if (KBD_ROWS_PIN & (1 << KBD_ROW3_IP)) // 3rd row
        scan[2] = 1;
    if (KBD_ROWS_PIN & (1 << KBD_ROW4_IP)) // 4th row
        scan[3] = 1;

    for (uint8_t i = 0; i < KBD_ROWS; i++)
    {
        if (scan[i])
            kbd[col][i]--;
        else
            kbd[col][i]++;

        if (kbd[col][i] < KBD_DEBOUNCE_MIN)
            kbd[col][i] = KBD_DEBOUNCE_MIN;
        else if (kbd[col][i] > KBD_DEBOUNCE_MAX)
            kbd[col][i] = KBD_DEBOUNCE_MAX;

        if (kbd[col][i] > KBD_DEBOUNCE_THRESH)
            kbdBitmap[KBD_BMP_BYTE(col)] |= KBD_BMP_MASK(col, i);
    }
}
