/******************************************************************************
  Copyright 2013 Tom Cornall

  This program is free software: you can redistribute it and/or modify
  it under the terms of the GNU General Public License as published by
  the Free Software Foundation, either version 3 of the License, or
  (at your option) any later version.

  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.

  You should have received a copy of the GNU General Public License
  along with this program.  If not, see <http://www.gnu.org/licenses/>.  
 ******************************************************************************/
#include "Frontend.h"

using namespace std;

/*
 *
 */
Frontend::Frontend(DiagInterface &diag, QWidget *parent):
    QWidget(parent),
    diag(diag)
{
    setWindowTitle("Beamspring USB Util");

    /*
     * vref stuff
     */
    vrefMaskTimer = new QTimer(this);
    vrefMaskTimer->setSingleShot(true);

    QGroupBox *vrefGroup = new QGroupBox("Voltage threshold");

    QLabel *vrefLabel = new QLabel("<p>The threshold that a key is considered "
            "pressed or not.</p>"
            "<p>If a valid set of scancodes is loaded (some keys "
            "must be set to PRESSED or RELEASED) then in most cases the "
            "threshold can be auto-calibrated when the keyboard "
            "is plugged in (this happens if 65535 is stored in the "
            "controller's EEPROM).</p>"
            "<p>Alternatively, a forced override value can "
            "be stored in the controller's EEPROM.</p>");
    vrefLabel->setWordWrap(true);

    QLabel *vrefSpinBoxLabel = new QLabel("Current threshold:");
    vrefSpinBox = new QSpinBox;
    vrefSpinBox->setMaximum(65535);
    QHBoxLayout *vrefSpinBoxHBox = new QHBoxLayout;
    vrefSpinBoxHBox->addWidget(vrefSpinBoxLabel);
    vrefSpinBoxHBox->addWidget(vrefSpinBox);

    autoCalButton = new QPushButton;
    storeVrefButton =
        new QPushButton("Store current threshold as override in EEPROM");

    QVBoxLayout *vrefVBox = new QVBoxLayout;
    vrefVBox->addWidget(vrefLabel);
    vrefVBox->addLayout(vrefSpinBoxHBox);
    vrefVBox->addWidget(autoCalButton);
    vrefVBox->addWidget(storeVrefButton);

    vrefGroup->setLayout(vrefVBox);

    /*
     * bootloader section
     */
    QGroupBox *bootloaderGroup = new QGroupBox("Bootloader");

    QLabel *bootloaderLabel = new QLabel("Enter the bootloader to allow "
            "Atmel Flip or dfu-programmer to flash new firmware.");
    bootloaderLabel->setWordWrap(true);

    bootloaderButton = new QPushButton("Enter bootloader");

    QVBoxLayout *bootloaderVBox = new QVBoxLayout;
    bootloaderVBox->addWidget(bootloaderLabel);
    bootloaderVBox->addWidget(bootloaderButton);

    bootloaderGroup->setLayout(bootloaderVBox);

    /*
     * matrix stuff
     */
    QTimer *keyStatesTimer = new QTimer(this);
    keyStatesTimer->start(0);

    QGroupBox *matrixGroup = new QGroupBox("Layout");

    layerConditionsGrid = new QGridLayout;

    matrixGrid = new QGridLayout;
    matrixGrid->setSpacing(0);

    QFrame *vline = new QFrame;
    vline->setFrameShape(QFrame::VLine);

    QHBoxLayout *matrixHBox = new QHBoxLayout;
    matrixHBox->addLayout(layerConditionsGrid);
    matrixHBox->addWidget(vline);
    matrixHBox->addLayout(matrixGrid);

    matrixWidget = new QWidget;
    matrixWidget->setLayout(matrixHBox);

    matrixScrollArea = new QScrollArea;
    matrixScrollArea->setWidget(matrixWidget);
    matrixScrollArea->setWidgetResizable(true);
    matrixScrollArea->setMinimumHeight(200);
    matrixScrollArea->setMinimumWidth(800);

#if QT_VERSION >= 0x050000
    QScroller::grabGesture(matrixScrollArea,
            QScroller::LeftMouseButtonGesture); // heh cool
#endif

    loadMatrixButton = new QPushButton("Load from EEPROM");
    storeMatrixButton = new QPushButton("Store in EEPROM");

    QPushButton *importMatrixButton = new QPushButton("Import layout");
    QPushButton *exportMatrixButton = new QPushButton("Export layout");

    QHBoxLayout *layoutHBox = new QHBoxLayout;
    layoutHBox->addWidget(loadMatrixButton);
    layoutHBox->addWidget(storeMatrixButton);
    layoutHBox->addWidget(importMatrixButton);
    layoutHBox->addWidget(exportMatrixButton);

    QVBoxLayout *matrixLayout = new QVBoxLayout;
    matrixLayout->addLayout(layoutHBox);
    matrixLayout->addWidget(matrixScrollArea);

    matrixGroup->setLayout(matrixLayout);

    /*
     * main layout
     */
    QVBoxLayout *vbox1 = new QVBoxLayout;
    vbox1->addWidget(vrefGroup);
    vbox1->addWidget(bootloaderGroup);
    vbox1->addWidget(matrixGroup);

    setLayout(vbox1);

    /* ask for a few things immediately */
    QTimer::singleShot(0, this, SLOT(updateVref()));
    QTimer::singleShot(0, this, SLOT(buildMatrix()));
    QTimer::singleShot(0, this, SLOT(buildLayerConditions()));

    connect(vrefSpinBox, SIGNAL(valueChanged(int)),
            SLOT(vrefValueChanged(int)));
    connect(vrefMaskTimer, SIGNAL(timeout()), SLOT(setVrefFromBox()));
    connect(autoCalButton, SIGNAL(clicked()), SLOT(autoCalButtonClicked()));
    connect(storeVrefButton, SIGNAL(clicked()), SLOT(storeVrefButtonClicked()));
    connect(bootloaderButton, SIGNAL(clicked()),
            SLOT(bootloaderButtonClicked()));
    connect(keyStatesTimer, SIGNAL(timeout()), SLOT(updateKeyStates()));
    connect(storeMatrixButton, SIGNAL(clicked()),
            SLOT(storeMatrixButtonClicked()));
    connect(loadMatrixButton, SIGNAL(clicked()),
            SLOT(loadMatrixButtonClicked()));
    connect(importMatrixButton, SIGNAL(clicked()),
            SLOT(importMatrixButtonClicked()));
    connect(exportMatrixButton, SIGNAL(clicked()),
            SLOT(exportMatrixButtonClicked()));
}

/*
 *
 */
Frontend::~Frontend(void)
{
}

/*
 *
 */
void Frontend::updateVref(void)
{
    disconnect(vrefSpinBox, SIGNAL(valueChanged(int)), this,
            SLOT(vrefValueChanged(int)));

    vrefSpinBox->setValue(diag.vref());

    connect(vrefSpinBox, SIGNAL(valueChanged(int)), this,
            SLOT(vrefValueChanged(int)));

    /* re-enable widgets and set autocal button's text---should probably use
     * state machine or something as this is not needed every time
     */
    setEnabled(true);
    autoCalButton->setText("Auto-calibrate");
}

/*
 *
 */
void Frontend::vrefValueChanged(int)
{
    /* mask for a bit before actually setting value in case just mashing one
     * of the arrows, or typing etc.
     */
    vrefMaskTimer->start(500);
    autoCalButton->setEnabled(false);
    storeVrefButton->setEnabled(false);
}

/*
 *
 */
void Frontend::setVrefFromBox(void)
{
    qDebug() << "setting vref";
    diag.setVref(vrefSpinBox->value());
    autoCalButton->setEnabled(true);
    storeVrefButton->setEnabled(true);
}

/*
 *
 */
void Frontend::autoCalButtonClicked(void)
{
    setEnabled(false);
    autoCalButton->setText("Auto-calibrating...");

    diag.autoCalibrate();

    /* keyboard will disappear for a bit while it does its thing */
    QTimer::singleShot(1000, this, SLOT(autoCalComplete()));
}

/*
 *
 */
void Frontend::autoCalComplete(void)
{
    unsigned short vref = diag.vref();

    QMessageBox::StandardButton result = QMessageBox::question(this,
            "Auto-calibration complete",
            "Auto-calibration finished, threshold is " +
            QString::number(vref) + ". Would you like auto-calibration to "
            "take place automatically every time the keyboard is powered up?",
            QMessageBox::Yes | QMessageBox::No);

    if (result == QMessageBox::Yes)
    {
        diag.setVref(65535);
        diag.storeVref();
        diag.setVref(vref);
    }

    updateVref();
}

/*
 *
 */
void Frontend::storeVrefButtonClicked(void)
{
    diag.storeVref();
}

/*
 *
 */
void Frontend::bootloaderButtonClicked(void)
{
    QMessageBox::StandardButton result = QMessageBox::question(this,
            "Enter bootloader",
            "Keyboard will stop responding until either it is unplugged then "
            "plugged back in, or the programming tool resets it. Really "
            "enter bootloader?", QMessageBox::Yes | QMessageBox::No);

    if (result == QMessageBox::Yes)
    {
        setEnabled(false);
        diag.enterBootloader();
        QMessageBox::information(this, "Rebooted into bootloader",
                "The keyboard is ready to be flashed. Will now exit.");
        QTimer::singleShot(0, this, SLOT(close()));
    }
}

/*
 *
 */
void Frontend::buildMatrix(void)
{
    vector<vector<vector<unsigned char>>> scancodes = diag.scancodes();

    /* clear existing matrix widgets */
    QLayoutItem *item;
    while ((item = matrixGrid->takeAt(0)) != NULL)
    {
        delete item->widget();
        delete item;
    }
    keyWidgets.clear();

    int layers = scancodes.size();
    if (layers == 0)
        throw runtime_error("error: keyboard reports matrix with no layers");
    int cols = scancodes[0].size();
    if (cols == 0)
        throw runtime_error("error: keyboard reports matrix with no columns");
    int rows = scancodes[0][0].size();

    keyWidgets = vector<vector<vector<Key *>>>(layers,
            vector<vector<Key *>>(cols, vector<Key *>(rows)));

    for (int layer = 0; layer < layers; layer++)
    {
        QLabel *layerLabel = new QLabel("<b>" + (layer == 0 ? "Base Layer" :
                    "Layer " + QString::number(layer)) +
                ": </b>");
        matrixGrid->addWidget(layerLabel, (layer * rows) + layer + 1, 0,
                layers, 1, Qt::AlignLeft | Qt::AlignVCenter);
        for (int col = 0; col < cols; col++)
        {
            if (layer == 0)
            {
                QLabel *colLabel = new QLabel("<b>" + QString::number(col + 1) +
                        "</b>");
                colLabel->setAlignment(Qt::AlignCenter);
                matrixGrid->addWidget(colLabel, 0, col + 1);
            }

            for (int row = 0; row < rows; row++)
            {
                Key *key = new Key(diag, layer, col, row,
                        scancodes[layer][col][row]);
                keyWidgets[layer][col][row] = key;
                matrixGrid->addWidget(key, (layer * rows) + layer + row + 1,
                        col + 1);
            }
        }

        if (layer < layers - 1)
        {
            QFrame *hline = new QFrame;
            hline->setFrameShape(QFrame::HLine);
            matrixGrid->addWidget(hline, ((layer + 1) * rows) + layer + 1,
                    0, 1, cols + 1);
        }
    }
}

/*
 *
 */
void Frontend::buildLayerConditions(void)
{
    vector<LayerCondition> cnds = diag.layerConditions();

    /* clear existing widgets */
    QLayoutItem *item;
    while ((item = layerConditionsGrid->takeAt(0)) != NULL)
    {
        delete item->widget();
        delete item;
    }

    for (size_t i = 0; i < layerConditionWatchers.size(); i++)
        delete layerConditionWatchers[i];
    layerConditionWatchers.clear();

    int count = cnds.size();
    if (count == 0)
        throw runtime_error("error: keyboard reports zero layer conditions");

    QLabel *titleLabel = new QLabel("<b>Layer Conditions</b>");
    layerConditionsGrid->addWidget(titleLabel, 0, 0, 1, 5);
    for (int i = 0; i < 3; i++)
    {
        QLabel *fnLabel = new QLabel("<b>Fn" + QString::number(i + 1) + "</b>");
        layerConditionsGrid->addWidget(fnLabel, 1, i + 1);
    }
    for (int i = 0; i < count; i++)
    {
        QLabel *rowLabel = new QLabel("<b>" + QString::number(i + 1) + "</b>");
        layerConditionsGrid->addWidget(rowLabel, i + 2, 0);

        QCheckBox *fn1Check = new QCheckBox;
        fn1Check->setChecked(cnds[i].fn1Set());

        QCheckBox *fn2Check = new QCheckBox;
        fn2Check->setChecked(cnds[i].fn2Set());

        QCheckBox *fn3Check = new QCheckBox;
        fn3Check->setChecked(cnds[i].fn3Set());

        QLabel *arrowLabel = new QLabel(QChar(0x2192));

        QComboBox *layerCombo = new QComboBox;
        for (int i = 0; i < 4; i++)
            layerCombo->addItem(i == 0 ? "Base Layer" : "Layer " +
                    QString::number(i), i);
        layerCombo->setCurrentIndex(cnds[i].layer());

        layerConditionsGrid->addWidget(fn1Check, i + 2, 1);
        layerConditionsGrid->addWidget(fn2Check, i + 2, 2);
        layerConditionsGrid->addWidget(fn3Check, i + 2, 3);
        layerConditionsGrid->addWidget(arrowLabel, i + 2, 4);
        layerConditionsGrid->addWidget(layerCombo, i + 2, 5);

        layerConditionWatchers.push_back(new LayerConditionWatcher(diag,
                    i, fn1Check, fn2Check, fn3Check, layerCombo, this));
    }

    layerConditionsGrid->setRowStretch(layerConditionsGrid->rowCount(), 1);
}

/*
 *
 */
void Frontend::updateKeyStates(void)
{
    if (!isEnabled())
        return;
    if (keyWidgets.size() == 0)
        return;

    vector<vector<bool>> states = diag.keyStates();

    int cols = states.size();
    if (cols == 0)
        throw runtime_error("error: keyboad reports key states with no "
                "columns");
    int rows = states[0].size();

    for (size_t layer = 0; layer < keyWidgets.size(); layer++)
        for (int col = 0; col < cols; col++)
            for (int row = 0; row < rows; row++)
                keyWidgets[layer][col][row]->setPressed(states[col][row]);
}

/*
 *
 */
void Frontend::storeMatrixButtonClicked(void)
{
    setEnabled(false);
    storeMatrixButton->setText("Writing to EEPROM...");

    diag.storeScancodes();

    /* keyboard will disappear for a bit while it does its thing */
    QTimer::singleShot(1000, this, SLOT(storeMatrixComplete()));
}

/*
 *
 */
void Frontend::storeMatrixComplete(void)
{
    setEnabled(true);
    storeMatrixButton->setText("Store layout in EEPROM");
}

/*
 *
 */
void Frontend::loadMatrixButtonClicked(void)
{
    setEnabled(false);
    loadMatrixButton->setText("Reading from EEPROM...");

    diag.loadScancodes();

    /* keyboard will disappear for a (shorter) bit while it does its thing */
    QTimer::singleShot(250, this, SLOT(loadMatrixComplete()));
}

/*
 *
 */
void Frontend::loadMatrixComplete(void)
{
    setEnabled(true);
    loadMatrixButton->setText("Load layout from EEPROM");

    buildMatrix();
}

/*
 *
 */
void Frontend::importMatrixButtonClicked(void)
{
    QString filename = QFileDialog::getOpenFileName(this, "Import layout",
            QString(), "Layout files (*.l)");
    if (filename == QString())
        return;

    QString errStr;
    vector<vector<vector<unsigned char>>> scancodes;
    vector<LayerCondition> layerConditions;
    if (!importLayout(filename, scancodes, layerConditions, errStr))
    {
        QMessageBox::critical(this, "Error",
                QString("Could not import layout: ") + errStr);
        return;
    }

    try
    {
        diag.setScancodes(scancodes);
        diag.setLayerConditions(layerConditions);
        buildMatrix();
        buildLayerConditions();
    }
    catch (std::exception &e)
    {
        QMessageBox::critical(this, "Error", QString::fromUtf8(e.what()));
    }
}

/*
 *
 */
void Frontend::exportMatrixButtonClicked(void)
{
    QString filename = QFileDialog::getSaveFileName(this, "Export layout",
            "layout.l", "Layout files (*.l)");
    if (filename == QString())
        return;

    QString errStr;
    if (!exportLayout(filename, keyWidgets, layerConditionWatchers, errStr))
        QMessageBox::critical(this, "Error",
                QString("Could not export layout: ") + errStr);
}
